from __future__ import annotations

import importlib

from qadence_protocols.types import qadence_available_protocols


def available_protocols() -> dict:
    """Return the available protocols."""

    qadence_protocols: dict = dict()

    for protocol in qadence_available_protocols:
        module = importlib.import_module(f"qadence_protocols.{protocol}.protocols")
        ProtocolCls = getattr(module, protocol.capitalize())
        qadence_protocols[protocol] = ProtocolCls

    return qadence_protocols



from __future__ import annotations

import itertools

import torch
from qadence import QuantumCircuit, QuantumModel, block_to_tensor, chain, kron
from qadence.noise.protocols import Noise
from qadence.operations import I, X, Z
from qadence.types import BackendName
from torch import tensor


def twirl_swap(n_qubits: int, twirl: tuple, samples_twirl: dict) -> dict:
    output = {}
    operand = "".join(map(str, [1 if i in twirl else 0 for i in range(n_qubits)]))
    for key in samples_twirl.keys():
        output["{0:b}".format(int(key, 2) ^ int(operand, 2))] = samples_twirl[key]

    return output


def compute_exp(n_qubits: int, samples_twirl: list, observable: list) -> tensor:
    out = 0
    o = block_to_tensor(
        kron(*[Z(i) if i in observable else I(i) for i in range(n_qubits)])
    ).squeeze()

    for sample_twirl in samples_twirl:
        shots = sum(sample_twirl.values())
        for key in sample_twirl.keys():
            out += o[int(key, 2), int(key, 2)] * sample_twirl[key] / shots
    return out


def mitigate(
    n_qubits: int,
    circuit: QuantumCircuit,
    backend: BackendName,
    noise: Noise,
    n_shots: int,
    observable: list,
    options: dict,
) -> tensor:
    """Corrects for readout errors on expectation values using all possible twirl operations.

    See Equation at the end of page 3
    See Page(2) in https://arxiv.org/pdf/2012.09738.pdf for implementation

    Args:
        noise: Specifies confusion matrix and default error probability
        observable: a list of pauli Z strings specified by index locations

    Returns:
        Mitigated output is returned
    """

    twirls = list(
        itertools.chain.from_iterable(
            [list(itertools.combinations(range(n_qubits), k)) for k in range(1, n_qubits)]
        )
    )

    # Generate samples for all twirls of circuit
    samples_twirl_num_list = []
    samples_twirl_den_list = []
    for twirl in twirls:
        layer = [X(i) for i in twirl]
        block_twirl = chain(circuit.block, kron(*layer))

        circ_twirl_num = QuantumCircuit(block_twirl.n_qubits, block_twirl)
        model_twirl_num = QuantumModel(circuit=circ_twirl_num, backend=backend)
        samples_twirl_num = model_twirl_num.sample(noise=noise, n_shots=n_shots)[0]
        samples_twirl_num_list.append(twirl_swap(n_qubits, twirl, samples_twirl_num))

        circ_twirl_den = QuantumCircuit(block_twirl.n_qubits, kron(*layer))
        model_twirl_den = QuantumModel(circuit=circ_twirl_den, backend=backend)
        samples_twirl_den = model_twirl_den.sample(noise=noise, n_shots=n_shots)[0]
        samples_twirl_den_list.append(twirl_swap(n_qubits, twirl, samples_twirl_den))

    expectation: float = 0

    for o in observable:
        expectation_num = compute_exp(n_qubits, samples_twirl_num_list, o)
        expectation_den = compute_exp(n_qubits, samples_twirl_den_list, o)
        expectation += expectation_num / expectation_den

    return torch.real(expectation)



from __future__ import annotations

import numpy as np
import torch
from qadence import BackendName, QuantumModel
from qadence.backends.api import backend_factory
from qadence.backends.pulser.backend import Backend
from qadence.blocks import block_to_tensor
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.analog import ConstantAnalogRotation, InteractionBlock
from qadence.circuit import QuantumCircuit
from qadence.noise import Noise
from qadence.operations import AnalogRot
from qadence.transpile import apply_fn_to_blocks
from qadence.utils import Endianness
from torch import Tensor

supported_noise_models = [Noise.DEPOLARIZING, Noise.DEPHASING]


def zne(noise_levels: Tensor, zne_datasets: list[list]) -> Tensor:
    poly_fits = []
    for dataset in zne_datasets:  # Looping over batched observables.
        batched_observable: list = []
        n_params = len(dataset[0])
        for p in range(n_params):  # Looping over the batched parameters.
            rearranged_dataset = [s[p] for s in dataset]
            # Polynomial fit function.
            poly_fit = np.poly1d(
                np.polyfit(noise_levels, rearranged_dataset, len(noise_levels) - 1)
            )
            # Return the zero-noise extrapolated value.
            batched_observable.append(poly_fit(0.0))
        poly_fits.append(batched_observable)

    return torch.tensor(poly_fits)


def pulse_experiment(
    backend: Backend,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor],
    noise: Noise,
    stretches: Tensor,
    endianness: Endianness,
    state: Tensor | None = None,
) -> Tensor:
    def mutate_params(block: AbstractBlock, stretch: float) -> AbstractBlock:
        """Closure to retrieve and stretch analog parameters."""
        # Check for stretchable analog block.
        if isinstance(block, (ConstantAnalogRotation, InteractionBlock)):
            stretched_duration = block.parameters.duration * stretch
            stretched_omega = block.parameters.omega / stretch
            stretched_delta = block.parameters.delta / stretch
            # The Hamiltonian scaling has no effect on the phase parameter
            phase = block.parameters.phase
            qubit_support = block.qubit_support
            return AnalogRot(
                duration=stretched_duration,
                omega=stretched_omega,
                delta=stretched_delta,
                phase=phase,
                qubit_support=qubit_support,
            )
        return block

    zne_datasets = []
    noisy_density_matrices: list = []
    for stretch in stretches:
        # FIXME: Iterating through the circuit for every stretch
        # and rebuilding the block leaves is inefficient.
        # Best to retrieve the parameters once
        # and rebuild the blocks.
        stre = stretch.item()
        block = apply_fn_to_blocks(circuit.block, mutate_params, stre)
        stretched_register = circuit.register.rescale_coords(stre)
        stretched_circuit = QuantumCircuit(stretched_register, block)
        conv_circuit = backend.circuit(stretched_circuit)
        noisy_density_matrices.append(
            # Contain a single experiment result for the stretch.
            backend.run_dm(
                conv_circuit,
                param_values=param_values,
                state=state,
                noise=noise,
                endianness=endianness,
            )[0]
        )
    # Convert observable to Numpy types compatible with QuTip simulations.
    # Matrices are flipped to match QuTip conventions.
    converted_observable = [np.flip(block_to_tensor(obs).numpy()) for obs in observable]
    # Create ZNE datasets by looping over batches.
    for observable in converted_observable:
        # Get expectation values at the end of the time serie [0,t]
        # at intervals of the sampling rate.
        zne_datasets.append(
            [
                [dm.expect(observable)[0][-1] for dm in density_matrices]
                for density_matrices in noisy_density_matrices
            ]
        )
    # Zero-noise extrapolate.
    extrapolated_exp_values = zne(
        noise_levels=stretches,
        zne_datasets=zne_datasets,
    )
    return extrapolated_exp_values


def noise_level_experiment(
    backend: Backend,
    circuit: QuantumCircuit,
    observable: list[AbstractBlock],
    param_values: dict[str, Tensor],
    noise: Noise,
    endianness: Endianness,
    state: Tensor | None = None,
) -> Tensor:
    noise_probs = noise.options.get("noise_probs")
    zne_datasets: list = []
    # Get noisy density matrices.
    conv_circuit = backend.circuit(circuit)
    noisy_density_matrices = backend.run_dm(
        conv_circuit, param_values=param_values, state=state, noise=noise, endianness=endianness
    )
    # Convert observable to Numpy types compatible with QuTip simulations.
    # Matrices are flipped to match QuTip conventions.
    converted_observable = [np.flip(block_to_tensor(obs).numpy()) for obs in observable]
    # Create ZNE datasets by looping over batches.
    for observable in converted_observable:
        # Get expectation values at the end of the time serie [0,t]
        # at intervals of the sampling rate.
        zne_datasets.append(
            [
                [dm.expect(observable)[0][-1] for dm in density_matrices]
                for density_matrices in noisy_density_matrices
            ]
        )
    # Zero-noise extrapolate.
    extrapolated_exp_values = zne(noise_levels=noise_probs, zne_datasets=zne_datasets)
    return extrapolated_exp_values


def analog_zne(
    model: QuantumModel,
    options: dict,
    noise: Noise,
    param_values: dict[str, Tensor],
    state: Tensor | None = None,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    if model._backend_name != BackendName.PULSER:
        raise ValueError("Only BackendName.PULSER supports analog simulations.")
    backend = backend_factory(backend=model._backend_name, diff_mode=None)
    stretches = options.get("stretches", None)
    if stretches is not None:
        extrapolated_exp_values = pulse_experiment(
            backend=backend,
            circuit=model._circuit.original,
            observable=[obs.original for obs in model._observable],
            param_values=param_values,
            noise=noise,
            stretches=stretches,
            endianness=endianness,
            state=state,
        )
    else:
        extrapolated_exp_values = noise_level_experiment(
            backend=backend,
            circuit=model._circuit.original,
            observable=[obs.original for obs in model._observable],
            param_values=param_values,
            noise=noise,
            endianness=endianness,
            state=state,
        )
    return extrapolated_exp_values


def mitigate(
    model: QuantumModel,
    options: dict,
    noise: Noise | None = None,
    param_values: dict[str, Tensor] = dict(),
) -> Tensor:
    if noise is None or noise.protocol not in supported_noise_models:
        if model._noise is None or model._noise.protocol not in supported_noise_models:
            raise ValueError(
                "A Noise.DEPOLARIZING or Noise.DEPHASING model must be provided"
                " either to .mitigate() or through the <class QuantumModel>."
            )
        noise = model._noise

    mitigation_zne = analog_zne(
        model=model, options=options, noise=noise, param_values=param_values
    )
    return mitigation_zne



from __future__ import annotations

from enum import Enum


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))  # type: ignore


class Protocols(StrEnum):
    """The available protocols for running experiments."""

    MITIGATIONS = "mitigations"
    """The Mitigations protocol."""
    # TODO: Placeholder for the measurements protocol.
    # MEASUREMENTS = "measurements"
    # """The Measurements protocol."""


qadence_available_protocols = Protocols.list()



from __future__ import annotations

import importlib
from dataclasses import dataclass
from functools import partial
from typing import Callable

PROTOCOL_TO_MODULE = {
    "twirl": "qadence_protocols.mitigations.twirl",
    "readout": "qadence_protocols.mitigations.readout",
    "zne": "qadence_protocols.mitigations.analog_zne",
}


@dataclass
class Mitigations:
    TWIRL = "twirl"
    READOUT = "readout"
    ANALOG_ZNE = "zne"

    def __init__(self, protocol: str, options: dict = dict()) -> None:
        self.protocol: str = protocol
        self.options: dict = options

    def mitigation(self) -> Callable:
        try:
            module = importlib.import_module(PROTOCOL_TO_MODULE[self.protocol])
        except KeyError:
            ImportError(f"The module for the protocol {self.protocol} is not implemented.")
        # Partially pass the options.
        return partial(getattr(module, "mitigate"), options=self.options)

    def _to_dict(self) -> dict:
        return {"protocol": self.protocol, "options": self.options}

    @classmethod
    def _from_dict(cls, d: dict) -> Mitigations | None:
        if d:
            return cls(d["protocol"], **d["options"])
        return None

    @classmethod
    def list(cls) -> list:
        return list(filter(lambda el: not el.startswith("__"), dir(cls)))
